/******************************************************************************
 *
 * AMDiS - Adaptive multidimensional simulations
 *
 * Copyright (C) 2013 Dresden University of Technology. All Rights Reserved.
 * Web: https://fusionforge.zih.tu-dresden.de/projects/amdis
 *
 * Authors: 
 * Simon Vey, Thomas Witkowski, Andreas Naumann, Simon Praetorius, et al.
 *
 * This file is provided AS IS with NO WARRANTY OF ANY KIND, INCLUDING THE
 * WARRANTY OF DESIGN, MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 *
 *
 * This file is part of AMDiS
 *
 * See also license.opensource.txt in the distribution.
 * 
 ******************************************************************************/



/** \file FirstOrderTerm.h */

#ifndef AMDIS_FIRST_ORDER_TERM_H
#define AMDIS_FIRST_ORDER_TERM_H

#include <vector>
#include "AMDiS_fwd.h"
#include "OperatorTerm.h"
#include "AbstractFunction.h"
#include "ElInfo.h"
#include "traits/size.hpp"
#include "MatrixVectorOperations.h"

namespace AMDiS {

  /**
   * \ingroup Assembler
   * 
   * \brief
   * Describes the first order terms: \f$ b \cdot \nabla u(\vec{x}) \f$
   */
  class FirstOrderTerm : public OperatorTerm
  {
  public:
    /// Constructor.
    FirstOrderTerm(int deg) 
      : OperatorTerm(deg) 
    {}

    /// Destructor.
    virtual ~FirstOrderTerm() {}

    /// Evaluation of \f$ \Lambda b \f$.
    virtual void getLb(const ElInfo *elInfo,
		       std::vector<mtl::dense_vector<double> >& result) const = 0;

    /// Implenetation of FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor)  
    {
      int dow = Global::getGeo(WORLD);

      if (num_rows(grdUhAtQP) > 0) {
	for (int iq = 0; iq < nPoints; iq++) {
	  double resultQP = 0.0;
	  for (int i = 0; i < dow; i++)
	    resultQP += grdUhAtQP[iq][i];
	  
	  result[iq] += resultQP * factor;
	}
      }
    }

  protected:
    /// Evaluation of \f$ \Lambda \cdot b\f$ if b contains the value 1.0
    /// in each component.
    inline void l1(const DimVec<WorldVector<double> >& Lambda,
		   mtl::dense_vector<double>& Lb,
		   double factor) const
    {
      const int dim = size(Lambda);

      for (int i = 0; i < dim; i++) {
	double val = 0.0;
      
	for (int j = 0; j < dimOfWorld; j++)
	  val += Lambda[i][j];

	Lb[i] += val * factor;
      }    
    }

    /// Evaluation of \f$ \Lambda \cdot b\f$.
    inline void lb(const DimVec<WorldVector<double> >& Lambda,
		   const WorldVector<double>& b,
		   mtl::dense_vector<double>& Lb,
		   double factor) const
    {
      const int dim = size(Lambda);

      for (int i = 0; i < dim; i++) {
//      double val = Lambda[i] * b;
//	double val = 0.0;
//	
//	for (int j = 0; j < dimOfWorld; j++)
//	  val += Lambda[i][j] * b[j];
//	
//	Lb[i] += val * factor;
        Lb[i] += (Lambda[i] * b) * factor;
      }    
    }

    /// Evaluation of \f$ \Lambda \cdot b\f$.
    inline void lb_one(const DimVec<WorldVector<double> >& Lambda,
		       mtl::dense_vector<double>& Lb,
		       double factor) const
    {
      const int dim = size(Lambda);

      for (int i = 0; i < dim; i++)
	Lb[i] += Lambda[i][bOne] * factor;
    }

  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b \cdot \nabla u(\vec{x}) \f$ with a vector b which 
   * only consits of entries equal to one.
   */
//   class Simple_FOT : public FirstOrderTerm
//   {
//   public:
//     /// Constructor.
//     Simple_FOT() 
//       : FirstOrderTerm(0) 
//     {}
// 
//     /// Implements FirstOrderTerm::getLb().
//     inline void getLb(const ElInfo *elInfo, 
// 		      vector<mtl::dense_vector<double> >& Lb) const 
//     {
//       const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
//       const int nPoints = static_cast<int>(Lb.size());
// 
//       for (int iq = 0; iq < nPoints; iq++)
// 	l1(grdLambda, Lb[iq], 1.0);
//     }
//   };

  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b \cdot \nabla u(\vec{x}) \f$ with a vector b which 
   * only consits of entries equal to one.
   */
  class Simple_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    Simple_FOT(double f = 1.0) 
      : FirstOrderTerm(0)
    {
      factor = new double;
      *factor = f;
    }

    /// Constructor.
    Simple_FOT(double *fptr) 
      : FirstOrderTerm(0), factor(fptr) 
    {}

    /// Implements FirstOrderTerm::getLb().
    inline void getLb(const ElInfo *elInfo,
		      std::vector<mtl::dense_vector<double> >& Lb) const
    {
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      const int nPoints = static_cast<int>(Lb.size());

      for (int iq = 0; iq < nPoints; iq++)
	l1(grdLambda, Lb[iq], (*factor));
    }

  private:
    /// Pointer to the factor.
    double *factor;
  };

  // for compatibility
  typedef Simple_FOT FactorSimple_FOT;
  
  /**
   * \ingroup Assembler
   *
   * \brief 
   * First order term: \f$ b \cdot \nabla u(\vec{x}) \f$ with a given vector b.
   */
  class Vector_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    Vector_FOT(WorldVector<double> *wv, double fac_ = 1.0) 
      : FirstOrderTerm(0), b(wv), fac(fac_)
    {}

    /// Constructor.
    Vector_FOT(int bIdx, double fac_ = 1.0) 
      : FirstOrderTerm(0), fac(fac_)
    {
      bOne = bIdx;
    }

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo, 
	       std::vector<mtl::dense_vector<double> >& Lb) const 
    {
      const DimVec<WorldVector<double> > &grdLambda = elInfo->getGrdLambda();
      const int nPoints = static_cast<int>(Lb.size());

      if (bOne > -1) {
	for (int iq = 0; iq < nPoints; iq++)
	  lb_one(grdLambda, Lb[iq], fac);
      } else {
	for (int iq = 0; iq < nPoints; iq++)
	  lb(grdLambda, *b, Lb[iq], fac);
      }
    }

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor)
    {
      if (num_rows(grdUhAtQP) > 0) {
	double factor_ = factor * fac;
	for (int iq = 0; iq < nPoints; iq++)
	  result[iq] += *b * grdUhAtQP[iq] * factor_;
      }
    }

  protected:
    /// Vector which is multiplied with \f$ \nabla u(\vec{x}) \f$
    WorldVector<double> *b;
    double fac;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * Simple_FOT multiplied with \f$ f(v(\vec{x})) \f$.
   */
  class VecAtQP_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VecAtQP_FOT(DOFVectorBase<double> *dv, 
		AbstractFunction<double, double> *af,
		WorldVector<double> *wv);

    /// Constructor.
    VecAtQP_FOT(DOFVectorBase<double> *dv,
		AbstractFunction<double, double> *af,
		int bIdx);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
    void initElement(const ElInfo* smallElInfo,
		     const ElInfo* largeElInfo,
		     SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// v at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Function f.
    AbstractFunction<double, double> *f;

    ///
    WorldVector<double> *b;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * Simple_FOT multiplied with \f$ f(\vec{x}) \f$.
   */
  class CoordsAtQP_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    CoordsAtQP_FOT(AbstractFunction<double, WorldVector<double> > *af)
      : FirstOrderTerm(af->getDegree()), g(af) 
    {}

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FistOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    /// Stores coordinates at quadrature points. Set in \ref initElement().
    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    /// Function avaluated at world coordinates.
    AbstractFunction<double, WorldVector<double> > *g;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * Vector_FOT multiplied with \f$ f(\vec{x}) \f$.
   */
  class VecCoordsAtQP_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VecCoordsAtQP_FOT(AbstractFunction<double, WorldVector<double> > *af,
		      WorldVector<double> *wv)
      : FirstOrderTerm(af->getDegree()), g(af), b(wv)
    {}

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FistOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    /// Stores coordinates at quadrature points. Set in \ref initElement().
    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    /// Function evaluated at world coordinates.
    AbstractFunction<double, WorldVector<double> > *g;

    /// Coefficient vector.
    WorldVector<double> *b;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b(\nabla v(\vec{x})) \cdot \nabla u(\vec{x})\f$.
   */
  class VectorGradient_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VectorGradient_FOT(DOFVectorBase<double> *dv,
		       AbstractFunction<WorldVector<double>, WorldVector<double> > *af);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    DOFVectorBase<double>* vec;

    /// Function for b.
    AbstractFunction<WorldVector<double>, WorldVector<double> > *f;

    /// Gradient of v at quadrature points.
    mtl::dense_vector<WorldVector<double> > gradAtQPs;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b(v(\vec{x})) \cdot \nabla u(\vec{x})\f$.
   */
  class VectorFct_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VectorFct_FOT(DOFVectorBase<double> *dv,
		  AbstractFunction<WorldVector<double>, double> *fct);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const; 

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec;

    /// Vector v at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    /// Function for b.
    AbstractFunction<WorldVector<double>, double> *vecFct;
  };

  /**
   * \ingroup Assembler
   *
   * \brief
   * 
   */
  class VecFctAtQP_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VecFctAtQP_FOT(AbstractFunction<WorldVector<double>, WorldVector<double> > *af)
      : FirstOrderTerm(af->getDegree()), g(af) 
    {}

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implements FistOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;

    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    /// Stores coordinates at quadrature points. Set in \ref initElement().
    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    /// Function avaluated at world coordinates.
    AbstractFunction<WorldVector<double>, WorldVector<double> > *g;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b(v(\vec{x}), \nabla w(\vec{x})) \cdot \nabla u(\vec{x})\f$.
   */
  class VecGrad_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor.
    VecGrad_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
		BinaryAbstractFunction<WorldVector<double>, double, WorldVector<double> > *fct);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);

    /// Implementation of \ref OperatorTerm::initElement() for multilpe meshes.
    void initElement(const ElInfo* smallElInfo,
		     const ElInfo* largeElInfo,
		     SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const; 
    
    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec1;
    DOFVectorBase<double>* vec2;
    
    /// Vector v at quadrature points.
    mtl::dense_vector<double> vecAtQPs;

    mtl::dense_vector<WorldVector<double> > gradAtQPs;
    
    /// Function for b.
    BinaryAbstractFunction<WorldVector<double>, double, WorldVector<double> > *vecFct;
  };


  /**
   * \ingroup Assembler
   *
   * \brief
   * First order term: \f$ b(\nabla v(\vec{x}), \nabla w(\vec{x})) \cdot \nabla u(\vec{x})\f$.
   */
  class FctGrad2_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor
    FctGrad2_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
		BinaryAbstractFunction<WorldVector<double>, WorldVector<double>, WorldVector<double> > *vecFct_)
      : FirstOrderTerm(vecFct_->getDegree()), vec1(dv1), vec2(dv2), vecFct(vecFct_) 
    {}

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const; 
    
    /// Implements FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
  protected:
    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec1;

    /// DOFVector to be evaluated at quadrature points.
    DOFVectorBase<double>* vec2;
    
    /// Gradient v at quadrature points.
    mtl::dense_vector<WorldVector<double> > grad1AtQPs;

    /// Gradient v at quadrature points.
    mtl::dense_vector<WorldVector<double> > grad2AtQPs;

    /// Function for b.
    BinaryAbstractFunction<WorldVector<double>, WorldVector<double>, WorldVector<double> > *vecFct;
  };


  class Vec2AndGradAtQP_FOT : public FirstOrderTerm
  {
  public:
    Vec2AndGradAtQP_FOT(DOFVectorBase<double> *dv1, 
			DOFVectorBase<double> *dv2,
			TertiaryAbstractFunction<double, double, double, WorldVector<double> > *f_, WorldVector<double> *b_);

    void initElement(const ElInfo* elInfo, 
		     SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    void getLb(const ElInfo *elInfo, 
	       std::vector<mtl::dense_vector<double> >& Lb) const;
    
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    DOFVectorBase<double> *vec1, *vec2;

    mtl::dense_vector<double> vec1AtQPs, vec2AtQPs;

    mtl::dense_vector<WorldVector<double> > gradAtQPs1;

    TertiaryAbstractFunction<double, double, double, WorldVector<double> > *f;

    WorldVector<double> *b;
  };


  class FctVecAtQP_FOT : public FirstOrderTerm
  {
  public:
    FctVecAtQP_FOT(DOFVectorBase<double> *dv,
		   BinaryAbstractFunction<double, WorldVector<double>, double> *f_,
		   WorldVector<double> *b_);
    
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;
    
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
  protected:
    DOFVectorBase<double>* vec;

    mtl::dense_vector<double> vecAtQPs;

    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    BinaryAbstractFunction<double, WorldVector<double>, double> *f;

    WorldVector<double> *b;
  };


  class Vec2AtQP_FOT : public FirstOrderTerm
  {
  public:
    Vec2AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
		 BinaryAbstractFunction<double, double, double> *af,
		 WorldVector<double> *b_);

    Vec2AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2,
		 BinaryAbstractFunction<double, double, double> *af,
		 int bIdx);
    
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;
    
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
  protected:
    DOFVectorBase<double> *vec1, *vec2;

    mtl::dense_vector<double> vec1AtQPs, vec2AtQPs;

    BinaryAbstractFunction<double, double, double> *f;

    WorldVector<double> *b;
  };


  class Vec3AtQP_FOT : public FirstOrderTerm
  {
  public:
    Vec3AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2, DOFVectorBase<double> *dv3,
		    TertiaryAbstractFunction<double, double, double, double> *f,
		    WorldVector<double> *b);
    
    Vec3AtQP_FOT(DOFVectorBase<double> *dv1, DOFVectorBase<double> *dv2, DOFVectorBase<double> *dv3,
		    TertiaryAbstractFunction<double, double, double, double> *f,
		    int b);
    
    void initElement(const ElInfo* elInfo, SubAssembler* subAssembler,
		     Quadrature *quad = NULL);
    
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& Lb) const;
    
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);
    
  protected:
    DOFVectorBase<double> *vec1, *vec2, *vec3;

    TertiaryAbstractFunction<double, double, double, double>* f;

    mtl::dense_vector<double> vec1AtQPs, vec2AtQPs, vec3AtQPs;

    WorldVector<double> *b;
  };
  
  // for compatibility
  typedef Vec3AtQP_FOT Vec3FctAtQP_FOT;


  class General_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor
    General_FOT(std::vector<DOFVectorBase<double>*> vecs,
		std::vector<DOFVectorBase<double>*> grads,
		TertiaryAbstractFunction<WorldVector<double>, 
		WorldVector<double>,
		std::vector<double>, 
		std::vector<WorldVector<double> > > *f);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo*, 
		     SubAssembler* ,
		     Quadrature *quad= NULL);

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& result) const;

    /// Implenetation of FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    std::vector<DOFVectorBase<double>*> vecs_; 

    std::vector<DOFVectorBase<double>*> grads_;

    TertiaryAbstractFunction<WorldVector<double>, 
			     WorldVector<double>,
			     std::vector<double>, 
			     std::vector<WorldVector<double> > > *f_;

    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    std::vector<mtl::dense_vector<double> > vecsAtQPs;

    std::vector<mtl::dense_vector<WorldVector<double> > > gradsAtQPs_;
  };


  class GeneralParametric_FOT : public FirstOrderTerm
  {
  public:
    /// Constructor
    GeneralParametric_FOT(std::vector<DOFVectorBase<double>*> vecs,
			  std::vector<DOFVectorBase<double>*> grads,
			  QuartAbstractFunction<WorldVector<double>, 
			  WorldVector<double>,
			  WorldVector<double>,
			  std::vector<double>, 
			  std::vector<WorldVector<double> > > *f);

    /// Implementation of \ref OperatorTerm::initElement().
    void initElement(const ElInfo*, SubAssembler*, Quadrature *quad = NULL);

    /// Implements FirstOrderTerm::getLb().
    void getLb(const ElInfo *elInfo,
	       std::vector<mtl::dense_vector<double> >& result) const;

    /// Implenetation of FirstOrderTerm::eval().
    void eval(int nPoints,
	      const mtl::dense_vector<double>& uhAtQP,
	      const mtl::dense_vector<WorldVector<double> >& grdUhAtQP,
	      const mtl::dense_vector<WorldMatrix<double> >& D2UhAtQP,
	      mtl::dense_vector<double>& result,
	      double factor);

  protected:
    std::vector<DOFVectorBase<double>*> vecs_; 

    std::vector<DOFVectorBase<double>*> grads_;

    QuartAbstractFunction<WorldVector<double>, 
			  WorldVector<double>,
			  WorldVector<double>,
			  std::vector<double>, 
			  std::vector<WorldVector<double> > > *f_;

    mtl::dense_vector<WorldVector<double> > coordsAtQPs;

    WorldVector<double> elementNormal_;

    std::vector<mtl::dense_vector<double> > vecsAtQPs;

    std::vector<mtl::dense_vector<WorldVector<double> > > gradsAtQPs_;
  };

}

#endif
